@@ -4,6 +4,8 @@
 from collections import *
 import gzip
 import random
+import math
+import tempfile
 
 # Third party library imports
 import pysam
@@ -29,11 +31,11 @@ def Reads_index (
     * input_fn
         Path to the bam file to index
     * skip_unmapped
-        Filter out unmapped reads
+        Do not include unmapped reads in index
     * skip_secondary
-        Filter out secondary alignment
+        Do not include secondary alignment in index
     * skip_supplementary
-        Filter out supplementary alignment
+        Do not include supplementary alignment in index
     """
     # Define logger
     logger = get_logger (name="Alignment_Reads_index", verbose=verbose, quiet=quiet)
@@ -49,27 +51,28 @@ def Reads_index (
     idx_fn = input_fn+".idx.gz"
     try:
         with pysam.AlignmentFile(input_fn) as bam, gzip.open(idx_fn, "wt") as idx, tqdm(unit=" Reads", disable=not progress) as pbar:
+            idx.write("read_id\tref_id\tpointer\tindex\n")
             try:
+                i=0
                 while True:
+
                     # Save pointer and read_id
                     p = bam.tell()
                     read = next(bam)
                     pbar.update()
 
-                    valid_read, read_status = _eval_read(
-                        read=read,
-                        skip_unmapped = skip_unmapped,
-                        skip_secondary = skip_secondary,
-                        skip_supplementary = skip_supplementary)
+                    valid_read, read_status = _eval_read(read=read, skip_unmapped = skip_unmapped, skip_secondary = skip_secondary, skip_supplementary = skip_supplementary)
 
                     if valid_read:
                         c["Reads retained"][read_status]+=1
                         c["Reads retained"]["total"]+=1
-                        idx.write("{}\t{}\n".format(read.query_name, p))
+                        idx.write("{}\t{}\t{}\t{}\n".format(read.query_name, read.reference_name, p, i))
                     else:
                         c["Reads discarded"][read_status]+=1
                         c["Reads discarded"]["total"]+=1
 
+                    i+=1
+
             except (StopIteration, KeyboardInterrupt):
                 pass
 
@@ -117,7 +120,7 @@ def Reads_sample (
     idx_fn = input_fn+".idx.gz"
     if not file_readable(idx_fn):
         logger.info("Read index not found. Creating one")
-        index_reads (input_fn, **kwargs)
+        Reads_index (input_fn, **kwargs)
 
     # Create the output folder if it does exist yet
     mkdir(output_folder, exist_ok=True)
@@ -126,13 +129,13 @@ def Reads_sample (
     logger.info("Load index")
     l = []
 
-    with gzip.open(idx_fn, "rt") as idx:
-        for line in tqdm (idx, desc="\tIndex", disable=not progress):
-            l.append (int(line.rstrip().split("\t")[1]))
+    with gzip.open(idx_fn, "rt") as idx_fp:
+        _= next(idx_fp)
+        for line in tqdm (idx_fp, desc="\tIndex", disable=not progress):
+            l.append (int(line.rstrip().split("\t")[2]))
 
     # Get random reads
     logger.info("Write sample reads")
-
     with pysam.AlignmentFile(input_fn) as bam_in:
 
         for sample_num in range(1, n_samples+1):
@@ -155,20 +158,161 @@ def Reads_sample (
             logger.info("Indexing output bam file")
             pysam.index(output_fn)
 
+def References_sample (
+    input_fn:[str],
+    output_fn:str="out.bam",
+    selected_reads_fn:str="select_ref.txt",
+    frac_reads:float=0.5,
+    min_reads_ref:int=30,
+    rand_seed:int=42,
+    sorting_threads:int=4,
+    verbose:bool=False,
+    quiet:bool=False,
+    progress:bool=False,
+    **kwargs):
+    """
+    Randomly sample reads per references according to a fraction of the reads mapped to this reference for one or
+    several files and write selected reads in a new bam file
+    * input_fn
+        Bam file path or directory containing bam files or list of files, or regex or list of regex.
+        It is quite flexible. All files need to be sorted and aligned to the same reference file.
+    * output_fn
+        Path to the output bam file (sorted and indexed)
+    * selected_reads_fn
+        Path to the output text file containing all the read id selected
+    * frac_reads
+        Fraction of reads mapped to sample for each reference
+    * min_reads_ref
+        Minimal read coverage per file and reference before sampling
+    * rand_seed
+        Seed to use for the pseudo randon generator. For non deterministic behaviour set to None
+    * sorting_threads
+        Number of threads to use for bam file sorting
+    """
+
+    # Define logger
+    logger = get_logger (name="Alignment_Ref_sample", verbose=verbose, quiet=quiet)
+    logger.warning("Running Alignment Ref_sample")
+
+    # Set random seed
+    random.seed(rand_seed)
+
+    # Create the output folders if they do not exist yet
+    for fn in [output_fn, selected_reads_fn]:
+        mkdir(os.path.dirname(fn), exist_ok=True)
+
+    # Load index file
+    logger.warning("Index files")
+
+    header = None
+    d = OrderedDict()
+    c =  Counter()
+    for fn in super_iglob(input_fn):
+        logger.info("Indexing alignment file {}".format(fn))
+
+        # Check Bam file compliance
+        _check_bam (fn)
+
+        try:
+            with pysam.AlignmentFile(fn) as bam, tqdm(desc="\tReading ", unit=" Reads", disable=not progress) as pbar:
+                # Collect header and verify that all files where aligned with the same reference and parameters
+                if header is None:
+                    header = bam.header
+                elif bam.header.references != header.references:
+                    raise pyBioToolsError("Bam Header from file {fn} is not consistent with previous files")
+
+                # Parse reads and save index
+                while True:
+                    loc = bam.tell()
+                    read = next(bam)
+                    pbar.update()
+
+                    if read.is_unmapped:
+                        c["unmapped reads"]+=1
+                    elif read.is_secondary:
+                        c["secondary reads"]+=1
+                    elif read.is_supplementary:
+                          c["supplementary reads"]+=1
+                    else:
+                        c["primary reads"]+=1
+                        ref_id = read.reference_name
+                        if not ref_id in d:
+                            d[ref_id] = OrderedDict()
+                        if not fn in d[ref_id]:
+                            d[ref_id][fn] = []
+                        d[ref_id][fn].append(loc)
+
+        except (StopIteration, KeyboardInterrupt):
+            pass
+
+    logger.info("Raw read counts summary")
+    log_dict(c, logger.info)
+
+    # Select reads per ref
+    logger.warning("Randomly pick reads per references")
+    select_loc_d = defaultdict(list)
+    c = Counter()
+    for ref_id, fn_d in d.items():
+
+        valid_cov = True
+        if min_reads_ref is not None:
+            for fn, loc_list in fn_d.items():
+                n_reads = len(loc_list)
+                if n_reads < min_reads_ref:
+                    valid_cov = False
+                    break
+
+        if valid_cov is True:
+            c["valid references"]+=1
+            for fn, loc_list in fn_d.items():
+                n_reads = len(loc_list)
+                c["valid reads"]+=n_reads
+                n_samples = math.ceil(len(loc_list)*frac_reads)
+                sample_loc_list = random.sample(loc_list, n_samples)
+                select_loc_d[fn].extend(sample_loc_list)
+                c["valid sampled reads"]+=n_samples
+
+        else:
+            c["low coverage references skipped"]+=1
+
+    logger.warning("Sample reads and write to output file")
+    with tempfile.NamedTemporaryFile() as temp_bam:
+        with pysam.AlignmentFile(temp_bam.name, "wb", header=header) as bam_out, open(selected_reads_fn, "w") as selected_reads_fp:
+            for fn, loc_list in select_loc_d.items():
+                loc_list = sorted(loc_list)
+                logger.info("Writing selected reads for bam file {}".format(fn))
+                with pysam.AlignmentFile(fn) as bam_in:
+                    for loc in tqdm(loc_list, desc="\tWriting ", unit=" Reads", disable=not progress):
+                        bam_in.seek(loc)
+                        read = next(bam_in)
+                        bam_out.write(read)
+                        selected_reads_fp.write("{}\n".format(read.query_name))
+
+        logger.info("Sort BAM File")
+        pysam.sort(temp_bam.name, "-o", output_fn, "-@", str(sorting_threads))
+
+    logger.info("Index sorted BAM File")
+    pysam.index(output_fn)
+
+    logger.info("Selected read counts summary")
+    log_dict(c, logger.info)
+
 #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~Filter~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
 def Filter (
     input_fn:str,
     output_fn:str,
+    selected_reads_fn:str=None,
     skip_unmapped:bool=False,
     skip_secondary:bool=False,
     skip_supplementary:bool=False,
+    index_reads:bool=False,
     orientation:str=".",
     min_read_len:int=0,
     min_align_len:int=0,
     min_mapq:int=0,
     min_freq_identity:float=0,
-    select_ref:[str]=[],
-    exclude_ref:[str]=[],
+    select_ref_fn:str=None,
+    exclude_ref_fn:str=None,
     verbose:bool=False,
     quiet:bool=False,
     progress:bool=False,
@@ -178,12 +322,16 @@ def Filter (
         Path to the bam file to filter
     * output_fn
         Path to the write filtered bam file
+    * selected_reads_fn
+        Optional file where to write ids of selected reads
     * skip_unmapped
         Filter out unmapped reads
     * skip_secondary
         Filter out secondary alignment
     * skip_supplementary
         Filter out supplementary alignment
+    * index_reads
+        Index bam file with both pysam and pybiotools reads_index
     * orientation
         Orientation of alignment on reference genome {"+","-" ,"."}
     * min_read_len
@@ -194,10 +342,10 @@ def Filter (
         Minimal mapping quality score (mapq)
     * min_freq_identity
         Minimal frequency of alignment identity [0 to 1]
-    * select_ref
-        List of references on which the reads have to be mapped.
-    * exclude_ref
-        List of references on which the reads should not be mapped.
+    * select_ref_fn
+        File containing a list of references on which the reads have to be mapped.
+    * exclude_ref_fn
+        File containing a list of references on which the reads should not be mapped.
     """
     # Define logger
     logger = get_logger (name="Alignment_Filter", verbose=verbose, quiet=quiet)
@@ -207,13 +355,36 @@ def Filter (
     logger.info("Checking input bam file")
     _check_bam (input_fn)
 
-    # Get random reads
+    # Import selected and excuded references list
+    select_ref=[]
+    if select_ref_fn:
+        logger.info("Loading selected reference file ids")
+        with open(select_ref_fn, "r") as fp:
+            for l in fp:
+                select_ref.append(l.strip())
+            logger.debug("Found {} references to select".format(len(select_ref)))
+    exclude_ref=[]
+    if exclude_ref_fn:
+        logger.info("Loading excluded reference file ids")
+        with open(exclude_ref_fn, "r") as fp:
+            for l in fp:
+                exclude_ref.append(l.strip())
+            logger.debug("Found {} references to exclude".format(len(exclude_ref)))
+
+    # Make output dir if needed
+    logger.debug("Make output directories")
+    mkdir(os.path.dirname(output_fn), exist_ok=True)
+    if selected_reads_fn:
+        mkdir(os.path.dirname(selected_reads_fn), exist_ok=True)
+
+    # Parse reads
     logger.info("Parsing reads")
     c = defaultdict (Counter)
+    selected_read_ids = set()
     try:
         with pysam.AlignmentFile(input_fn) as bam_in:
             with pysam.AlignmentFile(output_fn, "wb", header=bam_in.header) as bam_out:
-                for read in tqdm(bam_in, desc="\t", unit=" Reads", disable=not progress):
+                for read in tqdm(bam_in, desc="\tReading ", unit=" Reads", disable=not progress):
                     valid_read, read_status = _eval_read(
                         read = read,
                         skip_unmapped = skip_unmapped,
@@ -231,13 +402,23 @@ def Filter (
                         c["Reads retained"][read_status]+=1
                         c["Reads retained"]["total"]+=1
                         bam_out.write(read)
+                        selected_read_ids.add(read.query_name)
                     else:
                         c["Reads discarded"][read_status]+=1
                         c["Reads discarded"]["total"]+=1
 
-        # Check Bam readability and index
-        logger.info("Indexing output bam file")
-        pysam.index(output_fn)
+        if index_reads:
+            logger.info("Indexing output bam file with pysam")
+            pysam.index(output_fn)
+
+            logger.info("Indexing output bam file with pyBiotools")
+            Reads_index(output_fn)
+
+        if selected_reads_fn:
+            logger.info("Write selected read_id list")
+            with open (selected_reads_fn, "w") as fp:
+                for read_id in sorted(selected_read_ids):
+                    fp.write("{}\n".format(read_id))
 
     # Print read count summary
     finally:
